@@ -1,11 +1,15 @@
 import re
 import typing
+from collections import defaultdict
 from itertools import chain
 
 from sqlparse import tokens
-from sqlparse.sql import Token, Parenthesis, Comparison, IdentifierList, Identifier
+from sqlparse.sql import Token, Parenthesis, Comparison, IdentifierList, Identifier, Function
+
+from djongo.utils import encode_key
 
 from . import query
+from .functions import SQLFunc
 from .sql_tokens import SQLToken, SQLStatement, SQLComparison, SQLIdentifier
 from ..exceptions import SQLDecodeError
 
@@ -74,7 +78,7 @@ class _BinaryOp(_Op):
     def __init__(self, *args, token='prev_token', **kwargs):
         super().__init__(*args, **kwargs)
         tok = SQLToken.token2sql(getattr(self.statement, token), self.query)
-        if isinstance(tok, SQLIdentifier):
+        if isinstance(tok, (SQLIdentifier, SQLFunc)):
             self._field = tok.field
         elif isinstance(tok, SQLComparison):
             self._field = SQLIdentifier(tok._token.left, tok.query).field
@@ -92,6 +96,9 @@ class _BinaryOp(_Op):
 
 
 class _InNotInOp(_BinaryOp):
+    @property
+    def is_in(self):
+        return (isinstance(self, InOp) and not self.is_negated) or (isinstance(self, NotInOp) and self.is_negated)
 
     def _fill_in(self, token):
         self._in = []
@@ -110,54 +117,30 @@ class _InNotInOp(_BinaryOp):
                 self._in.append(None)
 
     def negate(self):
-        raise SQLDecodeError('Negating IN/NOT IN not supported')
+        self.is_negated = True
 
     def to_mongo(self):
-        raise NotImplementedError
-
-    def _to_mongo(self, op):
-        ## FIX: incorrectly resolving to $_nested_in for all ops instead of the relevant one only.
         if self.query.nested_query is not None and not self._in:
-            return {
-                '$expr': {
-                    op: ['$' + self._field, '$_nested_in']
-                }
-            }
-
+            expr = {'$in': [f'${self._field}', '$_nested_in']}
+            expr = {'$not': expr} if not self.is_in else expr
+            return {'$expr': expr}
         else:
-            return {self._field: {op: self._in}}
+            expr = {'$in': self._in}
+            expr = {'$not': expr} if not self.is_in else expr
+            return {self._field: expr}
 
 
 class NotInOp(_InNotInOp):
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(name='NOT IN', *args, **kwargs)
-        tok = self.statement.next()
-        if not tok.match(tokens.Keyword, 'IN'):
-            raise SQLDecodeError
-        self._fill_in(self.statement.next())
-
-    def to_mongo(self):
-        op = '$nin' if not self.is_negated else '$in'
-        return self._to_mongo(op)
-
-    def negate(self):
-        self.is_negated = True
+    def __init__(self, *args, token='prev_token', **kwargs):
+        super().__init__(name='NOT IN', *args, token=token, **kwargs)
+        self._fill_in(getattr(self.statement, token)[-1])
 
 
 class InOp(_InNotInOp):
-
     def __init__(self, *args, token='prev_token', **kwargs):
         super().__init__(name='IN', *args, token=token, **kwargs)
         self._fill_in(getattr(self.statement, token)[-1])
 
-    def to_mongo(self):
-        op = '$in' if not self.is_negated else '$nin'
-        return self._to_mongo(op)
-
-    def negate(self):
-        self.is_negated = True
-
 
 class LikeOp(_BinaryOp):
 
@@ -176,8 +159,8 @@ class LikeOp(_BinaryOp):
         if not isinstance(to_match, str):
             raise SQLDecodeError
 
-        to_match = to_match.replace('%', '.*')
-        self._regex = '^' + to_match + '$'
+        to_match = re.escape(to_match).replace('%', '.*')
+        self._regex = f'^{to_match}$'
 
     def to_mongo(self):
         return {self._field: {'$regex': self._regex}}
@@ -400,6 +383,9 @@ class _StatementParser:
         elif tok.match(tokens.Keyword, 'OR'):
             op = OrOp(**kw)
 
+        elif any(t.match(tokens.Comparison, 'NOT IN') for t in [tok, *getattr(tok, 'tokens', [])]):
+            op = NotInOp(**kw, token='current_token')
+
         elif any(t.match(tokens.Comparison, 'IN') for t in [tok, *getattr(tok, 'tokens', [])]):
             op = InOp(**kw, token='current_token')
 
@@ -425,6 +411,10 @@ class _StatementParser:
         elif tok.match(tokens.Keyword, 'IS'):
             op = IsOp(**kw)
 
+        elif isinstance(tok, Function) and statement.next_token.match(tokens.Keyword, 'IS'):
+            statement.skip(2)
+            op = IsOp(**kw)
+
         elif tok.value in JSON_OPERATORS:
             op = JSONOp(**kw)
 
@@ -470,12 +460,12 @@ class _StatementParser:
             link_op()
             if isinstance(op, CmpOp):
                 self._cmp_ops.append(op)
-            if not isinstance(op, (CmpOp, ParenthesisOp, ColOp, JSONOp)):
+            if not isinstance(op, (CmpOp, ParenthesisOp, ColOp)):
                 self._op_precedence(op)
             prev_op = op
 
         if prev_op.lhs is None:
-            if isinstance(prev_op, (CmpOp, ParenthesisOp, ColOp, JSONOp)):
+            if isinstance(prev_op, (CmpOp, ParenthesisOp, ColOp)):
                 self._ops.append(prev_op)
 
     def _op_precedence(self, operator: _Op):
@@ -499,9 +489,9 @@ class _StatementParser:
             raise SQLDecodeError
 
         op = None
-        while self._ops:
-            op = self._ops.pop(0)
+        for op in self._ops:
             op.evaluate()
+        self._ops.clear()
         self._op = op
 
 
@@ -541,18 +531,23 @@ class CmpOp(_Op):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self._identifier = SQLToken.token2sql(self.statement.left, self.query)
+        self._operator = OPERATOR_MAP[self.statement.token_next(0)[1].value]
 
         if isinstance(self.statement.right, Identifier):
-            raise SQLDecodeError('Join using WHERE not supported')
-
-        self._operator = OPERATOR_MAP[self.statement.token_next(0)[1].value]
-        index = re_index(self.statement.right.value)
-
-        self._constant = self.params[index] if index is not None else None
-        if isinstance(self._constant, dict):
-            self._field_ext, self._constant = next(iter(self._constant.items()))
+            # Comparing against field in same doc vs a constant.
+            self._right_identifier = SQLToken.token2sql(self.statement.right, self.query)
+            if self._right_identifier.table == self._identifier.table:
+                self._compare_fields_in_same_doc = True
+            else:
+                raise SQLDecodeError('Join using WHERE not supported')
         else:
-            self._field_ext = None
+            # Comparing with a constant
+            index = re_index(self.statement.right.value)
+            self._constant = self.params[index] if index is not None else None
+            if isinstance(self._constant, dict):
+                self._field_ext, self._constant = next(iter(self._constant.items()))
+            else:
+                self._field_ext = None
 
     def negate(self):
         self.is_negated = True
@@ -560,22 +555,38 @@ class CmpOp(_Op):
     def evaluate(self):
         pass
 
-    def to_mongo(self):
+    def _compare_fields_to_mongo(self):
+        """Comparison for fields in same document.
+        https://docs.mongodb.com/manual/reference/operator/query/expr/#behavior
+        """
+        l_field = self._identifier.field
+        r_field = self._right_identifier.field
+        if not self.is_negated:
+            return {'$expr': {self._operator: [f'${l_field}', f'${r_field}']}}
+        else:
+            return {'$expr': {'$not': {self._operator: [f'${l_field}', f'${r_field}']}}}
+
+    def _compare_constant_to_mongo(self):
         field = self._identifier.field
         if self._field_ext:
             field += '.' + self._field_ext
-
         if not self.is_negated:
             return {field: {self._operator: self._constant}}
         else:
             return {field: {'$not': {self._operator: self._constant}}}
 
+    def to_mongo(self):
+        if getattr(self, '_compare_fields_in_same_doc', False):
+            return self._compare_fields_to_mongo()
+        else:
+            return self._compare_constant_to_mongo()
 
-## Fix for boolean fields
+
+## Column operator - Fix for boolean fields, e.g. WHERE study.deleted
 class ColOp(_Op):
 
     def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+        super().__init__(name='COL', *args, **kwargs)
         self._identifier = SQLToken.token2sql(self.statement, self.query)
 
     def negate(self):
@@ -591,7 +602,7 @@ class ColOp(_Op):
 class JSONOp(_Op):
 
     def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
+        super().__init__(name='JSON', *args, **kwargs)
         self._identifier = self.statement.prev_token.value
         self._operator = self.statement.current_token.value
         index = re_index(self.statement.next_token.value)
@@ -600,6 +611,32 @@ class JSONOp(_Op):
     def negate(self):
         self.is_negated = True
 
+    def lookup_to_mongo(self):
+        # For $contains queries with lookups
+        if not isinstance(self._constant, dict):
+            raise SQLDecodeError(f"Invalid $contains element: {self._constant}")
+        final = defaultdict(dict)
+        for k, v in self._constant.items():
+            field = k.rsplit('__', 1)
+
+            if len(field) == 1:  # simple equality
+                final[field[0]] = v
+                continue
+
+            field, lookup = '.'.join(field[:-1]), field[-1]
+            '''
+            Supports mongo comparison operators: https://docs.mongodb.com/manual/reference/operator/query/
+            By this design, we expect the user to mix traditional ORM query with mongo operators for $contains.
+            Alternatively, we could also support django-based lookups
+            (https://docs.djangoproject.com/en/3.2/ref/models/querysets/#field-lookups)
+            but that would require more complicated transformations here.
+            '''
+            if lookup in ['in', 'nin', 'lt', 'lte', 'gt', 'gte', 'eq', 'ne']:
+                final[field][f'${lookup}'] = v
+            else:
+                raise SQLDecodeError(f'Lookup {lookup} not supported in $contains')
+        return dict(final)
+
     def to_mongo(self):
         field = '.'.join([f[1:-1] for f in self._identifier.split('.')[1:]])
 
@@ -609,21 +646,28 @@ class JSONOp(_Op):
 
         elif self._operator == '$contains':
             if isinstance(self._constant, dict):
-                return {f'{field}.{k}': self._constant[k] for k in self._constant}
+                # https://docs.mongodb.com/manual/tutorial/query-array-of-documents/#a-single-nested-document-meets-multiple-query-conditions-on-nested-fields
+                if self.is_negated:
+                    return {field: {'$not': {'$elemMatch': self.lookup_to_mongo()}}}
+                else:
+                    return {field: {'$elemMatch': self.lookup_to_mongo()}}
             elif isinstance(self._constant, list):
-                return {field: {'$all': self._constant}}
+                if self.is_negated:
+                    return {field: {'$not': {'$all': self._constant}}}
+                else:
+                    return {field: {'$all': self._constant}}
             else:
                 raise SQLDecodeError(f'Invalid params for $contains: {self._constant}')
 
         elif self._operator == '$has_key':
-            return {f'{field}.{self._constant}': {'$exists': not self.is_negated}}
+            return {f'{field}.{encode_key(self._constant)}': {'$exists': not self.is_negated}}
 
         elif self._operator == '$has_keys':
-            return {f'{field}.{const}': {'$exists': not self.is_negated} for const in self._constant}
+            return {f'{field}.{encode_key(const)}': {'$exists': not self.is_negated} for const in self._constant}
 
         elif self._operator == '$has_any_keys':
             return {
-                '$or': [{f'{field}.{const}': {'$exists': not self.is_negated}} for const in self._constant]
+                '$or': [{f'{field}.{encode_key(const)}': {'$exists': not self.is_negated}} for const in self._constant]
             }
 
         else:
@@ -642,6 +686,8 @@ OPERATOR_MAP = {
     '<=': '$lte',
 }
 OPERATOR_PRECEDENCE = {
+    'COL': 10,
+    'JSON': 9,
     'IS': 8,
     'BETWEEN': 7,
     'LIKE': 6,
